--[[ arjLib functions:

	 parameters in brackets ([parameter]) are optional.
	for all functions <target> will default to "target" except buffing functions ('ibuff','cbuff') which defaults to "player" if "target" is hostile.
	also, party spells ('pbuff','pres','pheal') will always return current target if valid.

-- cast / use functions
cact(action, [target]) - use action by name
cattack() - cast attack if not active and target in range
cinfo([channel], [target]) - say target info into a channel
citem(item, [target], [inv]) - use ineventory/bag item
cpet(spell, [toggle]) - cast pet spell
cshape([shape]) - cast shape/stance
ctop(spell, [target], [rank]) - cast top rank of spell
ctrack([track, ...]) - list of tracking spells to sequence

-- buff / debuff functions
ibuff([spell], [target], [index]) - is buff active?
cbuff(spell, [target], [max]) - cast buff if not active
idebuff([spell], [target], [index]) - is debuff active?
cdebuff(spell, [target], [max]) - cast debuff if not active

-- raid / party functions
pbuff(spell, [class]) - buff target & raid/party members
pdebuff(spell, [debuff]) - cast spell if target/raid/party members have debuff
pchat(text, [target]) - say in raid/party
pheal([spell], [thold], [max]) - heal lowest target/party member
pres([spell]) - cast spell on dead raid/party member

-- miscellaneous
rsay(text, [...]) - list of strings, one is randomly said

]]--


--[[ general stuff ]]--
--
-- version number
local version = "19"
-- spell rank levels to find max buff rank for target
local buffRanks = {
		-- druid buffs...
	[al_GIFT_OF_THE_WILD] = { 50,60 },
	[al_MARK_OF_THE_WILD] = { 1,10,20,30,40,50,60 },
	[al_THORNS] = { 6,14,24,34,44,54 },
	[al_REJUVENATION] = { 4,10,16,22,28,34,40,46,52,58,60 },
		-- mage buffs...
	[al_AMPLIFY_MAGIC] = { 18,30,42,54 },
	[al_ARCANE_BRILLIANCE] = { 56 },
	[al_ARCANE_INTELLECT] = { 1,14,28,42,56 },
	[al_DAMPEN_MAGIC] = { 12,24,36,48,60 },
		-- paladin buffs...
	[al_BLESSING_OF_MIGHT] = { 4,12,22,32,42,52,60 },
	[al_BLESSING_OF_PROTECTION] = { 10,24,38 },
	[al_BLESSING_OF_SACRIFICE] = { 46,54 },
	[al_BLESSING_OF_SANCTUARY] = { 30,40,50,60 },
	[al_BLESSING_OF_WISDOM] = { 14,24,34,44,54,60 },
		-- priest buffs...
	[al_DIVINE_SPIRIT] = { 40,42,54 },
	[al_POWER_WORD_FORTITUDE] = { 1,12,24,36,48,60 },
	[al_POWER_WORD_SHIELD]={ 6,12,18,24,30,36,42,48,54,60 },
	[al_PRAYER_OF_FORTITUDE] = { 48,60 },
	[al_RENEW] = { 8,14,20,26,32,38,44,50,56,60 },
	[al_SHADOW_PROTECTION] = { 30,42,56 },
		-- warlock buffs...
	[al_SHADOW_WARD] = { 32,42,52 },
}
-- max heal levels to avoid over-healing
local healRanks = {
		-- druid heals...
	[al_HEALING_TOUCH] = { 51,112,243,445,694,894,1120,1427,1796,2230,2677 },
		-- paladin heals...
	[al_FLASH_OF_LIGHT] = { 72,110,163,221,299,383 },
	[al_HOLY_LIGHT] = { 47,90,187,356,553,780,1053,1388,1770 },
	[al_HOLY_SHOCK] = { 220,301,395 },
		-- priest heals...
	[al_FLASH_HEAL] = { 237,314,393,478,616,764,958 },
	[al_GREATER_HEAL] = { 1353,1717,2147,2674,2922 },
	[al_HEAL] = { 341,571,856,1072 },
	[al_LESSER_HEAL] = { 56,85,157 },
		-- shaman heals...
	[al_CHAIN_HEAL] = { 368,465,629 },
	[al_HEALING_WAVE] = { 44,78,155,316,440,622,854,1167,1561 },
	[al_LESSER_HEALING_WAVE] = { 186,281,381,514,705,928 },
}
-- ambiguous spells that require direct targetting
local ambiguousSpells = {
	[al_DISPEL_MAGIC] = true,
}
-- buffs that can not be simultaneously applied
local sameBuff = {
		-- mage buffs
	[al_ARCANE_INTELLECT] = { [al_ARCANE_BRILLIANCE] = true },
		-- druid buffs
	[al_MARK_OF_THE_WILD] = { [al_GIFT_OF_THE_WILD] = true },
		-- priest buffs
	[al_DIVINE_SPIRIT] = { [al_PRAYER_OF_SPIRIT] = true },
	[al_POWER_WORD_FORTITUDE] = { [al_PRAYER_OF_FORTITUDE] = true },
	[al_SHADOW_PROTECTION] = { [al_PRAYER_OF_SHADOW_PROTECTION] = true },
}
-- instant cast spells
local instCast = {
	[al_INSTANT] = true,
	[al_INSTANT_CAST] = true,
}
-- silly pet action buttons don't tooltip correctly?
local petActions = {
	[1] = PET_ACTION_ATTACK,
	[2] = PET_ACTION_FOLLOW,
	[3] = PET_ACTION_WAIT,
	[8] = PET_MODE_AGGRESSIVE,
	[9] = PET_MODE_DEFENSIVE,
	[10] = PET_MODE_PASSIVE,
}
-- spellbook & action bar cache
local actionCache, spellCache = {}, {}


--[[ internal functions ]]--
--
-- debug stuff (ChatFrame2 is the combat log)
function al_msg(text)
	if(ChatFrame2) then ChatFrame2:AddMessage(text, 0, 0.8, 0.9) end
end
-- decides if we need to change targets or not (not is always better, doesn't inturrupt what we're doing).
local function al_target(spell, target)
	if(ambiguousSpells[spell]) then	-- ambiguous spells /always/ need to be targetted
		if(UnitExists(target) and not UnitIsUnit(target,"target")) then
			TargetUnit(target)
			return(true)
		end
	elseif(GetCVar("autoSelfCast") ~= "0") then	-- Blizz's autoSelfCast adds a little spice
		if(UnitExists(target) and not UnitIsUnit(target,"target") and not (UnitIsUnit(target,"player") and UnitCanAttack("player","target"))) then
			TargetUnit(target)
			return(true)
		end
	else	-- old-fashioned graceful redirect
		if(UnitExists(target) and not UnitIsUnit(target,"target") and UnitIsFriend("player","target")) then
			TargetUnit(target)
			return(true)
		end
	end

end


--[[ event functions ]]--
--
-- init & feedback for the XML load
function al_load()
	this:RegisterEvent("PLAYER_ENTERING_WORLD")
	this:RegisterEvent("PLAYER_LEAVING_WORLD")
	al_msg(":: arjLib version "..version.." loaded ::")
end
-- on event
function al_event()
	if(event=="ACTIONBAR_SLOT_CHANGED") then
		actionCache = {}
	elseif(event=="LEARNED_SPELL_IN_TAB") then
		spellCache = {}
	elseif(event=="PLAYER_ENTERING_WORLD") then
		this:RegisterEvent("ACTIONBAR_SLOT_CHANGED")
		this:RegisterEvent("LEARNED_SPELL_IN_TAB")
	elseif(event=="PLAYER_LEAVING_WORLD") then
		this:UnregisterEvent("ACTIONBAR_SLOT_CHANGED")
		this:UnregisterEvent("LEARNED_SPELL_IN_TAB")
	end
end


--[[ cast / use functions ]]--
--
-- cact(action, [target])
--	use <action> (name) on <target>.
--	returns true or nil.
function cact(action, target)
	if(not action or actionCache[action] == -1) then return end
	if(not target) then target="target" end
	local function doAction(actionID)
		local i, j = GetActionCooldown(actionID)
		if(i == 0 and j == 0) then
			local t = al_target(action, target)
			i = IsActionInRange(actionID)
			if((i == nil or i > 0) and IsUsableAction(actionID)) then
				UseAction(actionID)
				if(t) then TargetLastTarget() end
				arjTT:SetAction(actionID)
				if(instCast[arjTTTextLeft2:GetText()] or instCast[arjTTTextLeft3:GetText()]) then SpellStopCasting() end
				al_msg("action :: "..action.." ("..actionID..") :: on "..target)
				return(true)
			end
		end
	end
	if(actionCache[action]) then
		return(doAction(actionCache[action]))
	else
		for a = 1, 120 do
			if(HasAction(a)) then
				arjTT:SetAction(a)
				if(arjTTTextLeft1:GetText() == action) then
					actionCache[action] = a
					return(doAction(a))
				end
			end
		end
	end
	actionCache[action] = -1
end
--
-- cattack()
--	if there is no target then target closest enemy.
--	casts (melee) attack if not already attacking and target is close range.
--	returns true if attacking or attack has been cast.
function cattack()
	if(not UnitExists("target")) then TargetNearestEnemy() end
	if(UnitCanAttack("player","target") and CheckInteractDistance("target",3)) then
		if(not UnitAffectingCombat("player") or not PlayerFrame.inCombat) then AttackTarget() end
		return(true)
	end
end
--
-- cinfo([channel], [target])
--	report <target> info on <channel>.
--	<channel> defaults to LocalDefense.
--	<channel> needs to be a global channel (like "say" or "guild") or a number.
function cinfo(channel, target)
	if(not target) then target="target" end
	if(not UnitExists(target)) then return end
	if(not channel) then
		local y = false
		for i = 1, 10 do
			_, y = GetChannelName(i)
			if(y and string.find(y, al_LOCALDEFENSE)) then channel=i break end
		end
		if(not channel) then return end
	end
	local t, x, y= false, 0, 0
	if(UnitExists(target)) then
		x = UnitPVPName(target)
		if(x) then t = x else t = UnitName(target) end
		x = UnitLevel(target)
		y = UnitClassification(target)
		if(UnitPlayerControlled(target)) then
			if(x < 0) then
				x = UnitLevel("player") + 10
				t = t.." : ?? ("..x.."+)"
			else
				t = t.." : "..x
			end
		elseif(y == "normal") then
			if(x < 0) then
				x = UnitLevel("player") + 10
				t = t.." (npc) : ?? ("..x.."+)"
			else
				t = t.." (npc) : "..x
			end
		else
			if(x < 0) then
				x = UnitLevel("player") + 10
				t = t.." ("..y..") : ??"
			else
				t = t.." ("..y..") : "..x
			end
		end
		x = UnitRace(target)
		if(x) then t = t.." "..x end
		x = UnitClass(target)
		if(x) then t = t.." "..x end
		x = GetSubZoneText()
		if(x ~= "") then
			t = t.." : "..x
		else
			t = t.." : "..GetZoneText()
		end
		x, y = GetPlayerMapPosition(target)
		if(x > 0 and y > 0) then
			t = t.." ["..math.floor(x * 100)..","..math.floor(y * 100).."]"
		else
			x, y = GetPlayerMapPosition("player")
			if(x > 0 and y > 0) then t = t.." ["..math.floor(x * 100)..","..math.floor(y * 100).."]" end
		end
		if(UnitIsPVP(target)) then t = t.." : PvP" end
		if(tonumber(channel)) then
			SendChatMessage(t, "CHANNEL", GetDefaultLanguage("player"), channel)
		elseif(channel) then
			SendChatMessage(t, channel)
		end
	end
end
--
-- citem(item, [target], [inv])
--	looks through your worn items and bags (in that order) for <item> (name) and uses it on <target>.
--	if <inv> is true will only check worn items (to avoid equiping an item).
--	returns true or nil.
--	NOTE :: using an item with a vendor dialog open will sell the item, using an equipable item from a bag (such as a trinket) will equip the item.
function citem(item, target, inv)
	if(not item) then return end
	if(not target) then target="target" end
	local x, y = 0, 0
	for i = 0, 19 do
		if(GetInventoryItemTexture("player", i)) then
			arjTT:SetInventoryItem("player", i)
			if(arjTTTextLeft1:GetText() == item) then
				x, y = GetInventoryItemCooldown("player", i)
				if(x == 0 and y == 0) then
					local t = al_target(item, target)
					UseInventoryItem(i)
					if(t) then TargetLastTarget() end
					al_msg("use item :: "..item.." :: inv slot("..i..")")
					return(true)
				end
			end
		end
	end
	if(not inv) then
		for b = 0, 4 do
			for i = 1, GetContainerNumSlots(b) do
				if(GetContainerItemLink(b, i)) then
					arjTT:SetBagItem(b, i)
					if(arjTTTextLeft1:GetText() == item) then
						x, y = GetContainerItemCooldown(b, i)
						if(x == 0 and y == 0) then
							local t = al_target(item, target)
							UseContainerItem(b, i)
							if(t) then TargetLastTarget() end
							al_msg("use item :: "..item.." :: bag("..b..") slot("..i..")")
							return(true)
						end
					end
				end
			end
		end
	end
end
--
-- cpet(spell, [toggle])
--	activates pet <spell>.
--	if <toggle> is nil just activate once, if <toggle> is 0 then set autocast to off, if <toggle> is anything else set autocast to on.
--	setting <toggle> does not activate, only sets autocast.
function cpet(spell, toggle)
	local x = false
	for i = 1, NUM_PET_ACTION_SLOTS do
		if(petActions[i] == spell or GetPetActionInfo(i) == spell) then x=i break end
	end
	if(x) then
		if(toggle) then
			local _, _, _, _, _, y, z = GetPetActionInfo(x)
			if(toggle == 0) then
				if(y and z) then
					TogglePetAutocast(x)
					al_msg("petcast :: toggle "..spell.." ("..x..") off")
					return(true)
				end
			elseif(y and not z) then
				TogglePetAutocast(x)
				al_msg("petcast :: toggle "..spell.." ("..x..") on")
				return(true)
			end
		else
			local i, j = GetPetActionCooldown(x)
			if(i == 0 and j == 0) then
				CastPetAction(x)
				al_msg("petcast :: firing "..spell.." ("..x..")")
				return(true)
			end
		end
	end
end
--
-- cshape([shape])
--	triggers <shape> (or stance).
--	if <shape> is already current returns nil.
--	if <shape> not available (eg. druid in another shape) triggers current shape (back to caster).
--	if <shape> is nil returns current shape/stance name.
function cshape(shape)
	local a, t, j = false, false, false
	local function doShape(shapeID)
		local x, y = GetShapeshiftFormCooldown(shapeID)
		if(x == 0 and y == 0) then CastShapeshiftForm(shapeID) return(true) end
	end
	for i  = 1, GetNumShapeshiftForms() do
		arjTT:SetShapeshift(i)
		if(arjTTTextLeft1:GetText() == shape) then t = i end
		_, _, j = GetShapeshiftFormInfo(i)
		if(j) then a = i end
	end
	if(shape) then
		if(a and a == t) then return end
		if(t) then _, _, _, j = GetShapeshiftFormInfo(t) else j = nil end
		if(j) then return(doShape(t)) elseif(a) then return(doShape(a)) end
	elseif(a) then
		arjTT:SetShapeshift(a)
		return(arjTTTextLeft1:GetText())
	end
end
--
-- ctop(spell, [target], [rank])
--	cast <rank> of <spell> on <target>.
--	if <rank> is not found then cast top rank.
--	returns true (if cast attempted) or nil (if cast not attempted).
function ctop(spell, target, rank)
	if(not spell or spellCache[spell] == -1) then return end
	if(not target) then target="target" end
	local function doSpell(spellID)
		local i, j = GetSpellCooldown(spellID, BOOKTYPE_SPELL)
		if(i == 0 and j == 0) then
			local t = al_target(spell, target)
			CastSpell(spellID, BOOKTYPE_SPELL)
			if(t) then TargetLastTarget() end
			if(SpellIsTargeting()) then
				if(SpellCanTargetUnit(target)) then
					SpellTargetUnit(target)
				else
					SpellStopCasting()
					return
				end
			end
			arjTT:SetSpell(spellID, BOOKTYPE_SPELL)
			if(instCast[arjTTTextLeft2:GetText()] or instCast[arjTTTextLeft3:GetText()]) then SpellStopCasting() end
			local s, r = GetSpellName(spellID, BOOKTYPE_SPELL)
			al_msg("casting :: "..s.." ("..r..") :: on "..target)
			return(true)
		end
	end
	if(not spellCache[spell]) then
		spellCache[spell] = {}
	elseif(not rank and spellCache[spell]["top"]) then
		return(doSpell(spellCache[spell]["top"]))
	elseif(rank and spellCache[spell][rank]) then
		return(doSpell(spellCache[spell][rank]))
	end
	local i, s, r, id = 1, false, false, false
	repeat
		s, r = GetSpellName(i, BOOKTYPE_SPELL)
		if(s == spell) then
			id = i
			if(rank and string.find(r, rank)) then break end
		end
		i = i + 1
	until(not s)
	if(id) then
		if(rank) then spellCache[spell][rank]=id else spellCache[spell]["top"]=id end
		return(doSpell(id))
	end
	spellCache[spell] = -1
end
--
-- ctrack(track, [...])
--	<track> is comma-seperated list of tracking types.
--	returns ctop of the next <track> after the currently active tracking.
--	with no <track> indicated returns name of current track.
function ctrack(...)
	local t = false
	if(GetTrackingTexture()) then
		arjTT:SetTrackingSpell()
		t = arjTTTextLeft1:GetText()
	end
	if(arg.n == 0) then
		return(t)
	elseif(arg.n == 1 and arg[1] == t) then
		return
	end
	for i in arg do
		if(arg[i] == t) then
			if(arg.n <= i) then
				return(ctop(arg[1]))
			else
				return(ctop(arg[i + 1]))
			end
		end
	end
	return(ctop(arg[1]))
end


--[[ buff / debuff functions ]]--
--
-- ibuff([spell], [target], [index])
--	returns number of applications (stack) of buff <spell> on <target> or nil.
--	if <spell> is nil then returns true if there's any buffs on <target>.
--	if <index> is true then returns the index instead of the stack count.
function ibuff(spell, target, index)
	if(not target) then if(UnitIsFriend("player","target")) then target="target" else target="player" end end
	local function checkBuff(buff)
		local i = 1
		while(UnitBuff(target, i)) do
			arjTT:SetUnitBuff(target, i)
			if(arjTTTextLeft1:GetText()==buff) then return(i) else i=i+1 	end
		end
	end
	if(spell) then
		local b = checkBuff(spell)
		if(not b) then
			if(sameBuff[spell]) then
				for x in sameBuff[spell] do
					b = checkBuff(x)
					if(b) then break end
				end
			end
		end
		if(b) then
			if(index) then
				return(b)
			else
				local _, x = UnitBuff(target, b)
				return(x)
			end
		end
	else
		return(UnitBuff(target, 1))
	end
end
--
-- cbuff(spell, [target], [max])
--	cast top rank of <spell> the <target> can take if not already active.
--	<target> defaults to "player".
--	<max> = max number of buff stacks.
--	retuns ctop or nil.
function cbuff(spell, target, max)
	if(not spell or spellCache[spell] == -1) then return end
	if(not target) then if(UnitIsFriend("player","target")) then target="target" else target="player" end end
	local x = ibuff(spell, target)
	if(not x or (max and max > x)) then
		local r = false
		if(buffRanks[spell]) then
			local l = UnitLevel(target) + 10
			for i in buffRanks[spell] do if(l >= buffRanks[spell][i]) then r=i end end
			if(not r) then return end
		end
		return(ctop(spell, target, r))
	end
end
--
-- idebuff([spell], [target], [index])
--	returns number of applications (stack) of debuff <spell> on <taget> or nil.
--	if <spell> is nil then returns if any debuffs on <target>.
--	if <index> is true then returns the index instead of the stack count.
function idebuff(spell, target, index)
	if(not target) then target="target" end
	if(spell) then
		local i = 1
		while(UnitDebuff(target, i)) do
			arjTT:SetUnitDebuff(target, i)
			if(arjTTTextLeft1:GetText() == spell) then
				if(index) then
					return(i)
				else
					local _, x = UnitDebuff(target, i)
					return(x)
				end
			else
				i = i + 1
			end
		end
	else
		return(UnitDebuff(target, 1))
	end
end
--
-- cdebuff(spell, [target], [max])
--	cast debuff <spell> on <target> if not already active.
--	<max> = max number of debuff stacks.
--	returns ctop or nil.
function cdebuff(spell, target, max)
	if(not spell or spellCache[spell] == -1) then return end
	if(not target) then target="target" end
	local x = idebuff(spell,target)
	if(not x or (max and max > x)) then return(ctop(spell, target)) end
end


--[[ raid / party functions ]]--
--
-- pbuff(spell, [class])
--	use cbuff(<spell>) on target, player, and raid/party members.
--	<class> is a comma seperated list of classes that will be buffed, nil <class> means all classes are viable.
--	returns buff target name or nil.
function pbuff(spell, ...)
	if(not spell or spellCache[spell] == -1) then return end
	local function check(t)
		if(arg.n > 0) then
			local c = UnitClass(t)
			for i=1,arg.n do if(arg[i] == c) then c=nil break end end
			if(c) then return end
		end
		if(CheckInteractDistance(t, 4) and not UnitIsDeadOrGhost(t)) then
			return(cbuff(spell, t))
		end
	end
	if(UnitIsFriend("player","target")) then
		if(check("target")) then return(UnitName("target")) end
	end
	if(check("player")) then return(UnitName("player")) end
	for i=1,GetNumPartyMembers() do if(check("party"..i)) then return(UnitName("party"..i)) end end
	for i=1,GetNumRaidMembers() do if(check("raid"..i)) then return(UnitName("raid"..i)) end end
end
--
-- pdebuff(spell, [debuff])
--	<debuff> is the class of debuff (eg. "Magic" or "Disease") as it shows on Right1 of the tooltip /or/ a specific debuff spell name.
--	if <debuff> is nil, cast spell on target with /any/ debuffs.
--	returns target name or nil.
function pdebuff(spell, debuff)
	if(not spell or spellCache[spell] == -1) then return end
	local function check(target)
		if(not UnitIsDeadOrGhost(target)) then
			if(debuff) then
				local i = 1
				while(UnitDebuff(target, i)) do
					arjTT:SetUnitDebuff(target, i)
					if(arjTTTextLeft1:GetText() == debuff or (arjTTTextRight1:IsShown() and arjTTTextRight1:GetText() == debuff)) then
						al_msg("al :: found "..debuff.." on "..target.." ("..arjTTTextLeft1:GetText().." - "..arjTTTextRight1:GetText()..")")
						return(ctop(spell, target, nil, true))
					else
						i = i + 1
					end
				end
			else
				if(UnitDebuff(target,1)) then return(ctop(spell,target)) end
			end
		end
	end
	if(UnitIsFriend("player","target")) then if(check("target")) then return(UnitName("target")) end end
	if(check("player")) then return(UnitName("player")) end
	for i=1,GetNumPartyMembers() do
		if(CheckInteractDistance("party"..i,4) and check("party"..i)) then return(UnitName("party"..i)) end
	end
	for i=1,GetNumRaidMembers()do
		if(CheckInteractDistance("raid"..i,4) and check("raid"..i)) then return(UnitName("raid"..i)) end
	end
end
--
-- pchat(text, [target])
--	send <text> to raid/party chat or whisper to <target> (must be "target" or a proper name).
function pchat(text,target)
	if(not text) then return end
	if(target) then
		if(target == "target") then target = UnitName("target") end
		return(SendChatMessage(text, "WHISPER", GetDefaultLanguage("player"), target))
	else
		if(GetNumRaidMembers() > 0) then
			return(SendChatMessage(text, "raid"))
		elseif(GetNumPartyMembers() > 0) then
			return(SendChatMessage(text, "party"))
		end
	end
end
--
-- pheal([spell], [thold], [max])
--	cast <spell> on raid/party member (inc. player) with lowest health percentage, checks heal rank maximums to avoid overhealing.
--	if <spell> is nil only returns name of lowest health raid member.
--	<thold> is the threshhold deficit to be considered for healing as a decimal (eg. "0.60" to heal a target < 60%, defalts to "0.90").
--	<max> is the target heal percent (ie. "0.80" for 80% or "1.2" for 120%, defaults to <thold>).  
--	does a SpellStopCasting() if there are no valid targets (assumes your previous target was healed).
function pheal(spell, thold, max)
	if(spell and spellCache[spell] == -1) then return end
	if(not thold) then thold = 0.9 end
	if(not max) then max = thold end
	local target, health, rank = false, thold, false
	local function check(t)
		if(not UnitIsDeadOrGhost(t) and not ibuff(spell, t)) then
			local x = UnitHealth(t) / UnitHealthMax(t)
			if(x < health) then target, health = t, x end
		end
	end
	if(UnitIsFriend("player", "target")) then check("target") end
	if(not target) then
		check("player")
		local x = GetNumRaidMembers()
		if(x > 0) then
			for i=1,x do if(CheckInteractDistance("raid"..i,4)) then check("raid"..i) end end
		else
			for i=1,GetNumPartyMembers() do if(CheckInteractDistance("party"..i,4)) then check("party"..i) end end
		end
	end
	if(target) then
		if(spell) then
			al_msg("healing :: "..target.." (".. floor(health * 1000) / 10 .."%)")
			if(buffRanks[spell]) then
				if(cbuff(spell, target)) then return(UnitName(target)) end
			else
				local d = (UnitHealthMax(target) * max) - UnitHealth(target)
				if(healRanks[spell] and (UnitIsUnit(target,"player") or UnitPlayerOrPetInParty(target) or UnitPlayerOrPetInRaid(target))) then
					rank = 1
					for i in healRanks[spell] do if(d > healRanks[spell][i]) then rank=i end end
				end
				if(ctop(spell, target, rank)) then return(UnitName(target)) end
			end
		else
			return(target)
		end
	elseif(spell and CastingBarFrame.casting) then 
		SpellStopCasting()
	end
end
--
-- pres([spell])
--	cast <spell> on dead raid/party member with highest level.
--	will automatically target Priest, Paladin, Shaman, or Druid if in range.
--	returns target name or nil.
--	if <spell> is nil then just return unit name (ie. "party3", no work done).
--	does a SpellStopCasting() if there are no valid targets.
function pres(spell)
	if(spell and spellCache[spell] == -1) then return end
	local target, level = false, 0
	local function check(t)
		if(UnitIsDeadOrGhost(t)) then
			local c, l = UnitClass(t), UnitLevel(t)
			if(c == "Priest" or c == "Paladin" or c == "Shaman") then
				target, level = t, 99
			elseif(l > level) then
				target, level = t, l
			end
		end
	end
	if(UnitIsFriend("player","target")) then check("target") end
	if(not target) then
		local x = GetNumRaidMembers()
		if(x > 0) then
			for i=1,x do if(CheckInteractDistance("raid"..i, 4)) then check("raid"..i) end end
		else
			for i=1,GetNumPartyMembers() do if(CheckInteractDistance("party"..i, 4)) then check("party"..i) end end
		end
	end
	if(target) then
		if(spell) then
			al_msg("ressing :: "..target.." (lvl "..level..")")
			if(ctop(spell, target)) then return(UnitName(target)) end
		else
			return(target)
		end
	elseif(spell and CastingBarFrame.casting) then
		SpellStopCasting()
	end
end


--[[ miscellaneous ]]--
--
-- rsay([text, ...])
--	<text> is a comma seperated list of strings, one is selected randomly and put into the "say" channel.
--	returns the index of the string.
function rsay(...)
	if(arg.n > 0) then
		local x = math.random(1, arg.n)
		SendChatMessage(arg[x], "SAY")
		return(x)
	end
end
--
-- wsay(text, target)
--	whisper <text> to <target>.
--	meant for use with functions like 'pbuff' that return nil if nothig is done (therefore no whisper is sent).
function wsay(text, target)
	if(text and target) then
		return(SendChatMessage(text, "WHISPER", GetDefaultLanguage("player"), target))
	end
end
